/******************************************************************************
 *
 * Name: sdiobus.c
 * Project: Wireless LAN, Bus driver for SDIO interface
 * Version: $Revision: 1.1 $
 * Date: $Date: 2007/01/18 09:21:35 $
 * Purpose: Bus driver for SDIO interface
 *
 ******************************************************************************/

/******************************************************************************
 *
 * (C)Copyright 2005 Marvell
 *
 * THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF SYSKONNECT
 * The copyright notice above does not evidence any
 * actual or intended publication of such source code.
 *
 * This Module contains Proprietary Information of SysKonnect
 * and should be treated as Confidential.
 *
 * The information in this file is provided for the exclusive use of
 * the licensees of SysKonnect.
 * Such users have the right to use, modify, and incorporate this code
 * into products for purposes authorized by the license agreement
 * provided they include this notice and the associated copyright notice
 * with any such product.
 * The information in this file is provided "AS IS" without warranty.
 *
 ******************************************************************************/

/******************************************************************************
 *
 * History:
 *
 *	$Log: sdiobus.c,v $
 *	Revision 1.1  2007/01/18 09:21:35  pweber
 *	Put under CVS control
 *	
 *	Revision 1.2  2005/12/08 12:01:00  ebauer
 *	Driver ckecks AND update voltage registry value if not correct
 *	
 *	Revision 1.1  2005/10/07 08:43:47  jschmalz
 *	Put SDIO with FPGA under CVS control
 *	
 *	
 ******************************************************************************/
#ifndef _lint
static const char SysKonnectFileId[] = "@(#)" __FILE__ " (C) Marvell ";
#endif /* !_lint */

#include "diagvers.h"           /* diag version control */

#include "h/skdrv1st.h"
#include "h/skdrv2nd.h"

#ifndef WINNT
#include <linux/proc_fs.h>      // proc filesystem
#endif
#ifdef WINNT
#ifndef POOL_COLD_ALLOCATION    // Workaround for missing define in w2k header
#define POOL_COLD_ALLOCATION 256        // Note this cannot encode into the header.
#endif
#endif

#ifdef DBG
SK_U32 stdDebugLevel = DBG_DEFAULT;
#ifdef WINNT
LARGE_INTEGER t1Time, t1LastTime, t1CurTime;
LARGE_INTEGER t2Time, t2LastTime, t2CurTime;
LARGE_INTEGER t3Time, t3LastTime, t3CurTime;
LARGE_INTEGER t4Time, t4LastTime, t4CurTime;
LARGE_INTEGER t5Time, t5LastTime, t5CurTime;
LARGE_INTEGER t6Time, t6LastTime, t6CurTime;
#endif // #ifdef WINNT
#endif // #ifdef DBG

#define USE_DEBOUNCE_CARD_IN
                                                                                                                                                                        /*--- mmoser 12/21/2006 ---*/
#define MAX_DEBOUNCE_TIME 100

#ifdef WINNT
#ifdef ALLOC_PRAGMA
#pragma alloc_text (INIT, DriverEntry)
#pragma alloc_text (PAGE, SDIOBus_DriverUnload)
#pragma alloc_text (PAGE, SDIOBus_CreateClose)
#pragma alloc_text (PAGE, SDIOBus_IoCtl)
#endif // #ifdef ALLOC_PRAGMA

UNICODE_STRING RegistryPath = { 0, 0, 0 };
#else // #ifdef WINNT
MODULE_DESCRIPTION("SDIO Bus Driver Module");
MODULE_AUTHOR("Michael Moser (mmoser@syskonnect.de)");
MODULE_LICENSE("Dual BSD/GPL");

static int sdio_major = 0;

int mrvlsdio_open(struct inode *inode, struct file *filp);
int mrvlsdio_release(struct inode *inode, struct file *filp);
int SDIOBus_IoCtl(struct inode *inode, struct file *filp,
                  unsigned int cmd, unsigned long arg);

//
// Module parameters
//

static int bus_type = SDIO_BUS_TYPE;
static int bus_width = SDIO_1_BIT;
static int block_size = SD_BLOCK_SIZE;
static int clock_speed = 0;
static int gpi_0_mode = 1;      // edge triggered
static int gpi_0_level = 1;     // high->low    
static int gpi_1_mode = 1;      // edge triggered
static int gpi_1_level = 0;     // low->high     

//static int debug_flags  = 0;
static int debug_flags = 0x0c;

static int sdio_voltage = FPGA_POWER_REG_3_3_V;

DECLARE_WAIT_QUEUE_HEAD(add_remove_queue);

// kernel 2.6.x
module_param(bus_type, int, S_IRUGO);
module_param(bus_width, int, S_IRUGO);
module_param(block_size, int, S_IRUGO);
module_param(clock_speed, int, S_IRUGO);
module_param(debug_flags, int, S_IRUGO);
module_param(sdio_voltage, int, S_IRUGO);
module_param(gpi_0_mode, int, S_IRUGO); // 0 = level triggered       1 = edge triggered
module_param(gpi_0_level, int, S_IRUGO);        // 0 = low->high or low      1 = high->low or high
module_param(gpi_1_mode, int, S_IRUGO); // 0 = level triggered       1 = edge triggered    
module_param(gpi_1_level, int, S_IRUGO);        // 0 = low->high or low      1 = high->low or high 

// prototypes
// for filling the "/proc/mrvsdio" entry
int mrvlsdio_read_procmem(char *buf, char **start, off_t offset, int count,
                          int *eof, void *data);

#endif

//
// Define an Interface Guid to access the proprietary Marvell SDIO bus driver interface.
// This guid is used to identify a specific interface in IRP_MN_QUERY_INTERFACE
// handler.
//
// {1D5837EE-F656-46b4-A8DA-5484A25AB13A}
DEFINE_INIT_GUID(GUID_MRVL_SDIOBUS_INTERFACE_2_STANDARD,
                 0x1d5837ee, 0xf656, 0x46b4, 0xa8, 0xda, 0x54, 0x84, 0xa2,
                 0x5a, 0xb1, 0x3a);

//
// Define a Guid for SDIO bus type. This is returned in response to
// IRP_MN_QUERY_BUS_INTERFACE on PDO.
//
DEFINE_INIT_GUID(GUID_MRVL_SDIOBUS_BUS_TYPE,
                 0x43159826, 0xA758, 0x4aa2, 0x9F, 0x98, 0xAB, 0x22, 0x9C,
                 0x03, 0x1E, 0xD1);
// {43159826-A758-4aa2-9F98-AB229C031ED1}

#ifdef WINNT
/******************************************************************************
 *
 *  DriverEntry  -
 *
 * Description:
 *
 * Notes:
 *
 * Context:
 *
 * Returns:
 *
 */

NTSTATUS
DriverEntry(IN PDRIVER_OBJECT DriverObject, IN PUNICODE_STRING RegPath)
{
    INITDEBUG();
    DBGPRINT(DBG_LOAD, ("INIT - Enter DriverEntry\n"));
    DBGPRINT(DBG_ALL,
             ("*** %s - New SDIO-Bus Interface***\n", SK_SIGNON_MSG));
    DBGPRINT(DBG_ALL, ("*** Built on %s %s ***\n", __DATE__, __TIME__));

    //
    // Save a copy of *RegPath in driver global RegistryPath for future reference.
    //
    // UNICODE_NULL terminate the path so that we can safely use RegistryPath.Buffer
    //   as a PWSTR.
    //
    {
        USHORT size = RegPath->Length + sizeof(WCHAR);
        RegistryPath.Buffer =
            ExAllocatePool((PagedPool | POOL_COLD_ALLOCATION), size);

        if (NULL == RegistryPath.Buffer) {
            return STATUS_NO_MEMORY;
        }

        RegistryPath.Length = 0;
        RegistryPath.MaximumLength = size;
        RtlCopyUnicodeString(&RegistryPath, RegPath);
        RegistryPath.Buffer[size / sizeof(WCHAR) - 1] = UNICODE_NULL;

        DBGPRINT(DBG_LOAD, ("*** %S ***\n", RegistryPath));
    }

    //
    // Set entry points into the driver
    //
    DriverObject->MajorFunction[IRP_MJ_CREATE] = SDIOBus_CreateClose;
    DriverObject->MajorFunction[IRP_MJ_CLOSE] = SDIOBus_CreateClose;
    DriverObject->MajorFunction[IRP_MJ_PNP] = SDIOBus_PnP;
    DriverObject->MajorFunction[IRP_MJ_POWER] = SDIOBus_Power;
    DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL] = SDIOBus_IoCtl;
    DriverObject->MajorFunction[IRP_MJ_SYSTEM_CONTROL] =
        SDIOBus_SystemControl;
    DriverObject->DriverUnload = SDIOBus_DriverUnload;
    DriverObject->DriverExtension->AddDevice = SDIOBus_AddDevice;

    return STATUS_SUCCESS;
}

/******************************************************************************
 *
 *  SDIOBus_CreateClose  -
 *
 * Description:
 *		Some outside source is trying to create a file against us.
 *
 *   	If this is for the FDO (the bus itself) then the caller
 *   	is trying to open the proprietary connection to tell us
 *   	to enumerate or remove a device.
 *
 *   	If this is for the PDO (an object on the bus) then this
 *   	is a client that wishes to use the SDIO device.
 *
 * Notes:
 *
 * Context:
 *
 * Returns:
 *
 */
NTSTATUS
SDIOBus_CreateClose(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp)
{
    PIO_STACK_LOCATION irpStack;
    NTSTATUS status;
    PFDO_DEVICE_DATA fdoData;

    PAGED_CODE();

    fdoData = (PFDO_DEVICE_DATA) DeviceObject->DeviceExtension;

    status = STATUS_INVALID_DEVICE_REQUEST;
    Irp->IoStatus.Information = 0;

    SDIOBus_IncIoCount(fdoData);

    //
    // If it's not for the FDO. We don't allow create/close on PDO
    //
    if (fdoData->common.IsFDO) {
        //
        // Check to see whether the bus is removed
        //

        if (fdoData->common.DevicePnPState == Deleted) {
            status = STATUS_DELETE_PENDING;
        } else {
            irpStack = IoGetCurrentIrpStackLocation(Irp);

            switch (irpStack->MajorFunction) {
            case IRP_MJ_CREATE:
                {
                    DBGPRINT(DBG_LOAD, ("SDIOBus_CreateClose() : Create \n"));
                    status = STATUS_SUCCESS;
                    break;
                }
            case IRP_MJ_CLOSE:
                {
                    DBGPRINT(DBG_LOAD, ("SDIOBus_CreateClose() : Close \n"));
                    status = STATUS_SUCCESS;
                    break;
                }
            default:
                {
                    DBGPRINT(DBG_LOAD,
                             ("SDIOBus_CreateClose() : Unhandled major function : %d \n",
                              irpStack->MajorFunction));
                    break;
                }
            }
        }
    }

    Irp->IoStatus.Status = status;
    IoCompleteRequest(Irp, IO_NO_INCREMENT);
    SDIOBus_DecIoCount(fdoData);

    return status;
}

/******************************************************************************
 *
 *  SDIOBus_IoCtl  -
 *
 * Description:	 Handle user mode IOCTL requests.
 *
 * Notes:
 *
 * Context:
 *
 * Returns:
 *
 */
NTSTATUS
SDIOBus_IoCtl(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp)
{
    PIO_STACK_LOCATION irpStack;
    NTSTATUS status;
    ULONG inlen, outlen;
    PCOMMON_DEVICE_DATA pDev;
    PFDO_DEVICE_DATA fdoData;
    PVOID buffer;
    ULONG data, *pData;
    ULONG ulDataRead;
    ULONG cnt;
    SK_U8 ucValue, ucPowerValue;

    PAGED_CODE();

    fdoData = (PFDO_DEVICE_DATA) DeviceObject->DeviceExtension;

    //
    // We only take Device Control requests for the FDO.
    // That is the bus itself.
    //

    if (!fdoData->common.IsFDO) {
        //
        // These commands are only allowed to go to the FDO.
        //
        status = STATUS_INVALID_DEVICE_REQUEST;
        Irp->IoStatus.Status = status;
        IoCompleteRequest(Irp, IO_NO_INCREMENT);
        return status;
    }
    //
    // Check to see whether the bus is removed
    //

    if (fdoData->common.DevicePnPState == Deleted) {
        Irp->IoStatus.Status = status = STATUS_DELETE_PENDING;
        IoCompleteRequest(Irp, IO_NO_INCREMENT);
        return status;
    }

    SDIOBus_IncIoCount(fdoData);

    irpStack = IoGetCurrentIrpStackLocation(Irp);

    buffer = Irp->AssociatedIrp.SystemBuffer;
    inlen = irpStack->Parameters.DeviceIoControl.InputBufferLength;
    outlen = irpStack->Parameters.DeviceIoControl.OutputBufferLength;

    status = STATUS_SUCCESS;
    DBGPRINT(DBG_IOCTL,
             ("SDIOBus_IoCtl() : IOCTL-Code : %d \n",
              irpStack->Parameters.DeviceIoControl.IoControlCode));

    pDev = (PCOMMON_DEVICE_DATA) & fdoData->common;
    pData = &data;

    switch (irpStack->Parameters.DeviceIoControl.IoControlCode) {
    case IOCTL_SDIOBUSENUM_SET_VOLTAGE:
        memcpy(pData, buffer, sizeof(ULONG));
        DBGPRINT(DBG_LOAD, ("IOCTL_SDIOBUSENUM_SET_VOLTAGE: %x\n", data));
        ucPowerValue = (SK_U8) data;
        MEM_WRITE_UCHAR(pDev, SK_SLOT_0, FPGA_POWER_REG_DATA, ucPowerValue);
        MEM_WRITE_UCHAR(pDev, SK_SLOT_0, FPGA_POWER_REG_CMD,
                        FPGA_POWER_REG_CMD_START);
        DBGPRINT(DBG_LOAD, ("wait for stable voltage\n"));
        cnt = 0;
        do {
            MEM_READ_UCHAR(pDev, SK_SLOT_0, FPGA_POWER_REG_STATUS, &ucValue);
            DBGPRINT(DBG_LOAD, ("PowerRegulatorControl: 0x%x\n", ucValue));
        } while (++cnt < 10000 && (ucValue & FPGA_POWER_REG_STABLE) == 0);
        DBGPRINT(DBG_LOAD, ("IOCTL_SDIOBUSENUM_SET_VOLTAGE: cnt=%d\n", cnt));

        break;
    case IOCTL_SDIOBUSENUM_SET_BUSTYPE:
        DBGPRINT(DBG_LOAD, ("IOCTL_SDIOBUSENUM_SET_BUSTYPE\n"));
        break;
    case IOCTL_SDIOBUSENUM_SET_BUSWIDTH:
        DBGPRINT(DBG_LOAD, ("IOCTL_SDIOBUSENUM_SET_BUSWIDTH\n"));
        break;
    case IOCTL_SDIOBUSENUM_SET_CLKSPEED:
        DBGPRINT(DBG_LOAD, ("IOCTL_SDIOBUSENUM_SET_CLKSPEED\n"));
        break;
    case IOCTL_SDIOBUSENUM_GET_BOARDREV:
        DBGPRINT(DBG_LOAD, ("IOCTL_SDIOBUSENUM_GET_BOARDREV\n"));
        MEM_READ_ULONG(pDev, SK_SLOT_0, 0x210, &ulDataRead);
        memcpy(buffer, &ulDataRead, sizeof(ulDataRead));
        Irp->IoStatus.Information = 4;
        break;
    case IOCTL_SDIOBUSENUM_GET_JUMPER:
        DBGPRINT(DBG_LOAD, ("IOCTL_SDIOBUSENUM_GET_JUMPER\n"));
        MEM_READ_ULONG(pDev, SK_SLOT_0, 0x210, &ulDataRead);
        memcpy(buffer, &ulDataRead, sizeof(ulDataRead));
        Irp->IoStatus.Information = 4;
        break;
    default:
        DBGPRINT(DBG_LOAD, ("IOCTL DEFAULT\n"));
        status = STATUS_INVALID_PARAMETER;
        break;
    }

    Irp->IoStatus.Status = status;
    IoCompleteRequest(Irp, IO_NO_INCREMENT);
    SDIOBus_DecIoCount(fdoData);
    return status;
}

/******************************************************************************
 *
 *  SDIOBus_IoCtl  -
 *
 * Description:	 Clean up everything we did in driver entry.
 *
 * Notes:
 *
 * Context:
 *
 * Returns:
 *
 */
VOID
SDIOBus_DriverUnload(IN PDRIVER_OBJECT DriverObject)
{
    PAGED_CODE();

    DBGPRINT(DBG_LOAD, ("SDIOBus_DriverUnload()\n"));

    //
    // All the device objects should be gone.
    //

                                                                                                                                                                                                /*--- mmoser 13.09.2005 ---*/
//      ASSERT (NULL == DriverObject->DeviceObject);

    //
    // Here we free all the resources allocated in the DriverEntry
    //

    return;
}

/******************************************************************************
 *
 *  SDIOBus_IncIoCount  -
 *
 * Description:	 This routine increments the number of requests the device receives
 *
 * Notes:
 *
 * Context:
 *
 * Returns:
 *
 */
VOID
SDIOBus_IncIoCount(IN PFDO_DEVICE_DATA FdoData)
{
    SK_U32 result;

    result = InterlockedIncrement(&FdoData->OutstandingIO);

                                                                                                                                                                                                /*--- mmoser 13.09.2005 ---*/
//      ASSERT(result > 0);
    //
    // Need to clear StopEvent (when OutstandingIO bumps from 1 to 2)
    //
    if (result == 2) {
        //
        // We need to clear the event
        //
        KeClearEvent(&FdoData->StopEvent);
    }

    return;
}

/******************************************************************************
 *
 *  SDIOBus_DecIoCount  -
 *
 * Description:	 This routine decrements as it complete the request it receives
 *
 * Notes:
 *
 * Context:
 *
 * Returns:
 *
 */
VOID
SDIOBus_DecIoCount(IN PFDO_DEVICE_DATA FdoData)
{
    SK_U32 result;

    result = InterlockedDecrement(&FdoData->OutstandingIO);

    if (result == 1) {
        //
        // Set the stop event. Note that when this happens
        // (i.e. a transition from 2 to 1), the type of requests we
        // want to be processed are already held instead of being
        // passed away, so that we can't "miss" a request that
        // will appear between the decrement and the moment when
        // the value is actually used.
        //

        KeSetEvent(&FdoData->StopEvent, IO_NO_INCREMENT, FALSE);
    }

    if (result == 0) {
        //
        // The count is 1-biased, so it can be zero only if an
        // extra decrement is done when a remove Irp is received
        //
                                                                                                                                                                                                        /*--- mmoser 13.09.2005 ---*/
//              ASSERT(FdoData->common.DevicePnPState == Deleted);

        //
        // Set the remove event, so the device object can be deleted
        //

        KeSetEvent(&FdoData->RemoveEvent, IO_NO_INCREMENT, FALSE);
    }

    return;
}

/******************************************************************************
 *
 *  SDIOBus_SystemControl  -
 *
 * Description:
 *   We have just received a System Control IRP.
 *
 * Notes:
 *
 * Context:
 *
 * Returns:
 *
 */
NTSTATUS
SDIOBus_SystemControl(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp)
{
    PFDO_DEVICE_DATA fdoData;
    NTSTATUS status;
    PIO_STACK_LOCATION stack;
    PCOMMON_DEVICE_DATA commonData;

    DBGPRINT(DBG_LOAD, ("SDIOBus_SystemControl()\n"));

    PAGED_CODE();

    stack = IoGetCurrentIrpStackLocation(Irp);

    commonData = (PCOMMON_DEVICE_DATA) DeviceObject->DeviceExtension;

    //
    // Just complete the request with the current status
    //
    status = Irp->IoStatus.Status;
    IoCompleteRequest(Irp, IO_NO_INCREMENT);
    return status;
}

NTSTATUS
SDIO_RegGetDword(IN ULONG RelativeTo,
                 IN PWSTR Path,
                 IN PWSTR ParameterName, IN OUT PULONG ParameterValue)
{
    NTSTATUS status;
    RTL_QUERY_REGISTRY_TABLE paramTable[2];
    ULONG defaultValue;

    if ((NULL == Path) || (NULL == ParameterName) || (NULL == ParameterValue)) {
        return STATUS_INVALID_PARAMETER;
    }

    DBGPRINT(DBG_LOAD,
             ("SDIO_RegGetDword - RelativeTo= %x, Path=<%S>, ParameterName=<%S>\n",
              RelativeTo, Path, ParameterName));

    //
    // set up table entries for call to RtlQueryRegistryValues
    //
    // leave paramTable[1] as all zeros to terminate the table
    //
    // use original value as default value
    //
    // use RtlQueryRegistryValues to do the grunge work
    //
    RtlZeroMemory(paramTable, sizeof(paramTable));

    defaultValue = *ParameterValue;

    paramTable[0].Flags = RTL_QUERY_REGISTRY_DIRECT;
    paramTable[0].Name = ParameterName;
    paramTable[0].EntryContext = ParameterValue;
    paramTable[0].DefaultType = REG_DWORD;
    paramTable[0].DefaultData = &defaultValue;
    paramTable[0].DefaultLength = sizeof(ULONG);

    status = RtlQueryRegistryValues(RelativeTo | RTL_REGISTRY_OPTIONAL,
                                    Path, &paramTable[0], NULL, NULL);

    if (status != STATUS_SUCCESS) {
        DBGPRINT(DBG_LOAD,
                 ("SDIO_RegGetDword - RtlQueryRegistryValues FAILED w/status=%x\n",
                  status));
    }

    DBGPRINT(DBG_LOAD,
             ("SDIO_RegGetDword - post-query <%S> *ParameterValue = %x\n",
              ParameterName, *ParameterValue));

    return status;
}

NTSTATUS
SDIO_RegGetDeviceParameterDword(IN PDEVICE_OBJECT Pdo,
                                IN PWSTR ParameterName,
                                IN OUT PULONG ParameterValue)
{
    NTSTATUS status;
    HANDLE hKey;
    RTL_QUERY_REGISTRY_TABLE queryTable[2];
    ULONG defaultValue;

    PAGED_CODE();

    status = IoOpenDeviceRegistryKey(Pdo,
                                     PLUGPLAY_REGKEY_DEVICE |
                                     PLUGPLAY_REGKEY_CURRENT_HWPROFILE,
                                     KEY_READ, &hKey);

    if (status != STATUS_SUCCESS) {
        DBGPRINT(DBG_LOAD,
                 ("SDIO_RegGetDeviceParameterDword - RtlQueryRegistryValues FAILED w/status=%x\n",
                  status));
        return status;
    }

    defaultValue = *ParameterValue;

    RtlZeroMemory(&queryTable, sizeof(queryTable));

    queryTable[0].Flags = RTL_QUERY_REGISTRY_DIRECT;
    queryTable[0].Name = ParameterName;
    queryTable[0].EntryContext = ParameterValue;
    queryTable[0].DefaultType = REG_DWORD;
    queryTable[0].DefaultData = &defaultValue;
    queryTable[0].DefaultLength = sizeof(ULONG);

//              _asm int 3;

    status =
        RtlQueryRegistryValues(RTL_REGISTRY_HANDLE, hKey, queryTable, NULL,
                               NULL);

    if (status != STATUS_SUCCESS) {
        *ParameterValue = defaultValue;
        DBGPRINT(DBG_LOAD,
                 ("SDIO_RegGetDeviceParameterDword - RtlQueryRegistryValues FAILED w/status=%x\n",
                  status));
    }

    DBGPRINT(DBG_LOAD,
             ("SDIO_RegGetDeviceParameterDword - post-query <%S> *ParameterValue = %x\n",
              ParameterName, *ParameterValue));

    ZwClose(hKey);

    return status;
}

NTSTATUS
SDIO_RegSetDword(IN ULONG RelativeTo,
                 IN PCWSTR Path,
                 IN PCWSTR ValueName,
                 IN PVOID ValueData, IN ULONG ValueLength)
{
    NTSTATUS status;
    ULONG defaultValue;

    PAGED_CODE();

    if ((NULL == Path) || (NULL == ValueName) || (NULL == ValueData)) {
        return STATUS_INVALID_PARAMETER;
    }

    status = RtlWriteRegistryValue(RelativeTo | RTL_REGISTRY_OPTIONAL,
                                   Path,
                                   ValueName,
                                   REG_DWORD, ValueData, ValueLength);

    if (status != STATUS_SUCCESS) {
        DBGPRINT(DBG_LOAD, ("SDIO_RegSetDword Error"));
    }

    return status;

}

#else // #ifdef WINNT
static PFDO_DEVICE_DATA fdo_data = NULL;

static int SDIOThread(void *data);

static int probe(struct pci_dev *dev, const struct pci_device_id *id);
static void remove(struct pci_dev *dev);

// Call back functions for busdriver module

static sdio_add_card_notifier sdio_add_card_callback = NULL;
static sdio_remove_card_notifier sdio_remove_card_callback = NULL;

// 2.6 : task queue mechanism
struct work_struct sdiobus_wq;

static DECLARE_COMPLETION(on_exit);

// kernel 2.6

static struct pci_device_id ids[] = {
    {PCI_DEVICE(PCI_VENDOR_ID_SYSKONNECT, 0x8000),},
    {0,}
};

MODULE_DEVICE_TABLE(pci, ids);

/*
 * The fops
 */

struct file_operations mrvlsdio_fops = {
  llseek:NULL,
  read:NULL,
  write:NULL,
  ioctl:SDIOBus_IoCtl,
  mmap:NULL,
  open:mrvlsdio_open,
  release:mrvlsdio_release,
};

/*
 * Open and close
 */

int
mrvlsdio_open(struct inode *inode, struct file *filp)
{
    ENTER();

    if (fdo_data == NULL) {
        DBGPRINT(DBG_LOAD,
                 (KERN_DEBUG
                  "mrvlsdio_open: fdo_data==NULL -> Adapter not in CardBus slot!\n"));
        return -ENODEV;
    }
    filp->f_op = &mrvlsdio_fops;
    /* and use filp->private_data to point to the device data */
    filp->private_data = fdo_data;
    DBGPRINT(DBG_LOAD,
             (KERN_DEBUG
              "mrvlsdio_open: fdo_data=0x%p fdo_data.common=0x%p fdo_data.common.IOBase[0]= 0x%p\n",
              fdo_data, &fdo_data->common, fdo_data->common.IOBase[0]));

    LEAVE();
    return 0;                   /* success */
}

int
mrvlsdio_release(struct inode *inode, struct file *filp)
{
    ENTER();
    LEAVE();
    return 0;
}

/*
 * The ioctl() implementation
 */

int
SDIOBus_IoCtl(struct inode *inode, struct file *filp,
              unsigned int cmd, unsigned long arg)
{
    int ret = 0;
    UCHAR ucValue;
    ULONG ulValue;
    ULONG cnt;
    PFDO_DEVICE_DATA fdo;

    ENTER();

    fdo = (FDO_DEVICE_DATA *) filp->private_data;
    if (fdo == NULL) {
        DBGPRINT(DBG_LOAD,
                 (KERN_DEBUG "SDIOBus_IoCtl : fdo= 0x%p !!!!!!!!!!!!!!!!\n",
                  fdo));
        return -ENOTTY;
    }

    DBGPRINT(DBG_LOAD,
             (KERN_DEBUG
              "SDIOBus_IoCtl : fdo= 0x%p fdo->common.IOBase[0]= 0x%p\n", fdo,
              fdo->common.IOBase[0]));

    switch (cmd) {
    case IOCTL_SDIOBUSENUM_SET_VOLTAGE:
        {
            ret = get_user(ucValue, (int *) arg);
            DBGPRINT(DBG_LOAD,
                     (KERN_DEBUG
                      "SDIOBus_IoCtl : IOCTL_SDIOBUSENUM_SET_VOLTAGE: 0x%2.02X\n",
                      ucValue));
            MEM_WRITE_UCHAR(&fdo->common, SK_SLOT_0, FPGA_POWER_REG_DATA,
                            ucValue);
            MEM_WRITE_UCHAR(&fdo->common, SK_SLOT_0, FPGA_POWER_REG_CMD,
                            FPGA_POWER_REG_CMD_START);
            DBGPRINT(DBG_LOAD, (KERN_DEBUG "wait for stable voltage\n"));
            cnt = 0;
            do {
                MEM_READ_UCHAR(&fdo->common, SK_SLOT_0, FPGA_POWER_REG_STATUS,
                               &ucValue);
                DBGPRINT(DBG_LOAD,
                         (KERN_DEBUG "PowerRegulatorControl: 0x%x\n",
                          ucValue));
            } while (++cnt < 10000 && (ucValue & FPGA_POWER_REG_STABLE) == 0);
            DBGPRINT(DBG_LOAD,
                     ("IOCTL_SDIOBUSENUM_SET_VOLTAGE: cnt=%d\n", cnt));
            break;
        }
    case IOCTL_SDIOBUSENUM_GET_BOARDREV:
        {
            MEM_READ_ULONG(&fdo->common, SK_SLOT_0, FPGA_CARD_REVISION,
                           &ulValue);

            ret = put_user(ulValue, (int *) arg);
            DBGPRINT(DBG_LOAD,
                     (KERN_DEBUG
                      "SDIOBus_IoCtl : IOCTL_SDIOBUSENUM_GET_BOARDREV 0x%8.08X\n",
                      ulValue));
            break;
        }
    case IOCTL_SDIOBUSENUM_GET_JUMPER:
        {
            DBGPRINT(DBG_LOAD, ("IOCTL_SDIOBUSENUM_GET_JUMPER\n"));
            MEM_READ_ULONG(&fdo->common, SK_SLOT_0, FPGA_CARD_REVISION,
                           &ulValue);
            ret = put_user(ulValue, (int *) arg);
            DBGPRINT(DBG_LOAD,
                     (KERN_DEBUG
                      "SDIOBus_IoCtl : IOCTL_SDIOBUSENUM_GET_JUMPER 0x%8.08X\n",
                      ulValue));
            break;
        }

    case IOCTL_SDIOBUSENUM_SET_BUSTYPE:
        {
            DBGPRINT(DBG_LOAD,
                     (KERN_DEBUG
                      "SDIOBus_IoCtl : IOCTL_SDIOBUSENUM_SET_BUSTYPE --> IGNORED!\n"));
            break;
        }
    case IOCTL_SDIOBUSENUM_SET_BUSWIDTH:
        {
            DBGPRINT(DBG_LOAD,
                     (KERN_DEBUG
                      "SDIOBus_IoCtl : IOCTL_SDIOBUSENUM_SET_BUSWIDTH --> IGNORED!\n"));
            break;
        }
    case IOCTL_SDIOBUSENUM_SET_CLKSPEED:
        {
            DBGPRINT(DBG_LOAD,
                     (KERN_DEBUG
                      "SDIOBus_IoCtl : IOCTL_SDIOBUSENUM_SET_CLKSPEED --> IGNORED!\n"));
            break;
        }
    default:
        {
            DBGPRINT(DBG_LOAD,
                     (KERN_DEBUG
                      "SDIOBus_IoCtl : invalid cmd=%d (0x%8.08X)!\n", cmd,
                      cmd));
            return -ENOTTY;
        }
    }

    LEAVE();
    return ret;

}

static struct pci_driver mrvlsdio = {
    .name = "mrvlsdio",
    .id_table = ids,
    .probe = probe,
    .remove = remove,
};

static int
probe(struct pci_dev *dev, const struct pci_device_id *id)
{
    int i, ret;
    SK_U8 ucValue;
    SK_U32 mem;
    SK_U32 ulValue;

    DBGPRINT(DBG_LOAD, (KERN_DEBUG "mrvlsdio : probe()\n"));

    pci_enable_device(dev);
    if (pci_set_dma_mask(dev, 0xffffffff)) {
        DBGPRINT(DBG_LOAD,
                 (KERN_DEBUG "mrvlsdio: 32-bit PCI DMA not supported"));
        pci_disable_device(dev);
        return -EIO;
    }
    pci_set_master(dev);
                                                                                                                                                                                /*--- mmoser 4/4/2006 ---*/
    if (fdo_data == NULL) {
        fdo_data =
            (PFDO_DEVICE_DATA) kmalloc(sizeof(FDO_DEVICE_DATA), GFP_KERNEL);

        if (fdo_data == NULL) {
            return -ENOMEM;
        }
        memset(fdo_data, 0, sizeof(FDO_DEVICE_DATA));
    }

    fdo_data->common.IOBaseIx = 0;
                                                                                                                                                                                /*--- mmoser 10/11/2006 ---*/
    fdo_data->common.dev = dev;
    pci_request_regions(dev, "MRVL-SDIO");

    for (i = 0; i < 5; i++) {
        mem = pci_resource_start(dev, i);

        if (mem != 0) {
            if (IORESOURCE_MEM == pci_resource_flags(dev, i)) {
                fdo_data->common.IOBaseLength[fdo_data->common.IOBaseIx] =
                    pci_resource_end(dev, i) - mem + 1;
                fdo_data->common.IOBase[fdo_data->common.IOBaseIx] =
                    ioremap_nocache(mem,
                                    fdo_data->common.IOBaseLength[fdo_data->
                                                                  common.
                                                                  IOBaseIx]);
                DBGPRINT(DBG_LOAD,
                         (KERN_DEBUG
                          "mrvlsdio: IOBase[%d] = 0x%8.08X IOBaseLength = %u\n",
                          i,
                          (SK_U32) fdo_data->common.IOBase[fdo_data->common.
                                                           IOBaseIx],
                          fdo_data->common.IOBaseLength[fdo_data->common.
                                                        IOBaseIx]));
                fdo_data->common.IOBaseIx++;
                if (fdo_data->common.IOBaseIx >= 2) {
                    DBGPRINT(DBG_LOAD,
                             (KERN_DEBUG
                              "mrvlsdio : Too many memory address ranges!\n"));
                    fdo_data->common.IOBaseIx = 0;
                }
            }
        }
    }

    // Dummy read. Function return SUCCESS but value is messy !!!
    pci_read_config_byte(dev, PCI_INTERRUPT_LINE,
                         (SK_U8 *) & fdo_data->common.Interrupt);

    for (i = 0; i < 3; i++) {
        ret =
            pci_read_config_byte(dev, PCI_INTERRUPT_LINE,
                                 (SK_U8 *) & fdo_data->common.Interrupt);
        if (ret) {
            DBGPRINT(DBG_LOAD,
                     (KERN_DEBUG
                      "mrvlsdio: Failed to retrieve Interrupt Line (%d) %d ret=%d\n",
                      i, fdo_data->common.Interrupt, ret));
            UDELAY(100);
            continue;
        } else {
            break;
            if (fdo_data->common.Interrupt > 13) {
                DBGPRINT(DBG_ALL,
                         ("mrvlsdio: common.Interrupt = %d\n",
                          fdo_data->common.Interrupt));
                if (fdo_data->common.Interrupt == 16) {
                    /* Dell Dimension 3000 & Dell Inspiron 6000 */
                    break;
                }
                UDELAY(100);
                continue;
            }
            break;
        }
    }
    if (i >= 3) {
        DBGPRINT(DBG_LOAD,
                 (KERN_DEBUG
                  "mrvlsdio: Failed to retrieve Interrupt Line (%d) ret=%d\n",
                  i, ret));
        return -ENODEV;
    }
    DBGPRINT(DBG_LOAD,
             (KERN_DEBUG "mrvlsdio: Interrupt Line = %d  (%d)\n",
              fdo_data->common.Interrupt, i));

    MEM_READ_UCHAR(&fdo_data->common, SK_SLOT_0, FPGA_CARD_REVISION,
                   &ucValue);
    DBGPRINT(DBG_LOAD,
             (KERN_DEBUG "mrvlsdio: CardRevision: 0x%X (fdo_data=0x%p)\n",
              ucValue, &fdo_data));

    init_waitqueue_head(&fdo_data->common.wq);

    fdo_data->common.bus_type = bus_type;
    fdo_data->common.bus_width = bus_width;
    fdo_data->common.ClockSpeed = clock_speed;
    fdo_data->common.debug_flags = debug_flags;
    fdo_data->common.blockSize = block_size;
    fdo_data->common.sdio_voltage = sdio_voltage;

    pci_read_config_byte(dev, 0x0c, &ucValue);
    if (ucValue == 0)
        pci_write_config_byte(dev, 0x0c, 2);
    pci_read_config_byte(dev, 0x0c, &ucValue);
    printk("config[0x0c]=0x%2.02X\n", ucValue);
    pci_read_config_byte(dev, 0x0d, &ucValue);
    printk("config[0x0d]=0x%2.02X\n", ucValue);
    printk("debug_flags=0x%8.08X\n", debug_flags);

    DBGPRINT(DBG_LOAD,
             (KERN_DEBUG "mrvlsdio: bus_type     : SDIO%s\n",
              bus_type == SDIO_BUS_TYPE ? "" : "/SPI"));
    DBGPRINT(DBG_LOAD,
             (KERN_DEBUG "mrvlsdio: bus_width    : %s\n",
              bus_width == SDIO_4_BIT ? "4-bit" : "1-bit"));
    DBGPRINT(DBG_LOAD,
             (KERN_DEBUG "mrvlsdio: block_size   : %d\n", block_size));
    DBGPRINT(DBG_LOAD,
             (KERN_DEBUG "mrvlsdio: clock_speed  : %d\n", clock_speed));
    DBGPRINT(DBG_LOAD,
             (KERN_DEBUG "mrvlsdio: debug_flags  : 0x%8.08X\n", debug_flags));
    DBGPRINT(DBG_LOAD,
             (KERN_DEBUG "mrvlsdio: sdio_voltage : %d\n", sdio_voltage));

                                                                                                                                                                        /*--- mmoser 3/12/2007 ---*/
    DBGPRINT(DBG_LOAD,
             (KERN_DEBUG "mrvlsdio: gpi_0_mode   : %s triggered\n",
              gpi_0_mode == 0 ? "level" : "edge"));
    DBGPRINT(DBG_LOAD,
             (KERN_DEBUG "mrvlsdio: gpi_0_level  : %s\n",
              gpi_0_level == 0 ? (gpi_0_mode ==
                                  0 ? "low" : "low->high") : (gpi_0_mode ==
                                                              0 ? "high" :
                                                              "high->low")));
    DBGPRINT(DBG_LOAD,
             (KERN_DEBUG "mrvlsdio: gpi_1_mode   : %s triggered\n",
              gpi_1_mode == 0 ? "level" : "edge"));
    DBGPRINT(DBG_LOAD,
             (KERN_DEBUG "mrvlsdio: gpi_1_level  : %s\n",
              gpi_1_level == 0 ? (gpi_1_mode ==
                                  0 ? "low" : "low->high") : (gpi_1_mode ==
                                                              0 ? "high" :
                                                              "high->low")));

    // mmoser 2005-11-29
    atomic_set(&fdo_data->common.card_add_event, 0);
    atomic_set(&fdo_data->common.card_remove_event, 0);

                                                                                                                                                                        /*--- mmoser 8/29/2006 ---*/
    init_waitqueue_head(&fdo_data->common.trans_complete_evt.wq);
    init_waitqueue_head(&fdo_data->common.cmd_complete_evt.wq);

                                                                                                                                                                        /*--- mmoser 3/12/2007 ---*/
    MEM_READ_ULONG(&fdo_data->common, SK_SLOT_0, 0x200, &ulValue);
    ulValue &= 0x0FFFFFFF;
    ulValue |=
        ((gpi_1_mode << 3 | gpi_1_level << 2 | gpi_0_mode << 1 | gpi_0_level)
         << 28);
    printk("mrvlsdio: ulValue = 0x%8.08X\n", ulValue);
    MEM_WRITE_ULONG(&fdo_data->common, SK_SLOT_0, 0x200, ulValue);

    printk("mrvlsdio: reg[0x200] = 0x%8.08X\n", ulValue);

    SDHost_Init(&fdo_data->common);

    // mmoser 2005-11-29
    if (0 != request_irq(fdo_data->common.Interrupt,
                         SDIOBus_Isr, SA_SHIRQ,
                         MRVL_DEVICE_NAME, (PVOID) fdo_data)
        ) {
        DBGPRINT(DBG_LOAD,
                 (KERN_DEBUG
                  "mrvlsdio: Failed to register interrupt handler\n"));
        return -ENODEV;
    }

    if (SDHost_isCardPresent(&fdo_data->common) == SK_TRUE) {
        if (SDIOBus_CardPluggedIn(fdo_data) == SK_TRUE) {
            fdo_data->CardIn = SK_TRUE;
        }
    }
// mmoser 2005-11-22 start the thread as the last action

// mmoser 2005-11-21
    fdo_data->common.stop_thread = SK_FALSE;

    fdo_data->common.thread_id = kernel_thread(SDIOThread,
                                               fdo_data,
                                               (CLONE_FS | CLONE_FILES |
                                                CLONE_SIGHAND));
#ifdef SYSKT_DMA_MALIGN_TEST
    fdo_data->common.dma_tx_malign = 0;
    fdo_data->common.dma_rx_malign = 0;
    // fdo_data->common.dma_start_malign = 4096 - 32;
    fdo_data->common.dma_start_malign = 0;
#endif

    if (fdo_data->common.thread_id == 0)
        DBGPRINT(DBG_LOAD,
                 (KERN_DEBUG "mrvlsdio: Failed to start kernel thread!\n"));

    return 0;
}

static void
remove(struct pci_dev *dev)
{
    int j;
    DBGPRINT(DBG_LOAD, (KERN_DEBUG "remove mrvlsdio ... \n"));

    /* clean up any allocated resources and stuff here.
     * like call release_region();
     */
    if (fdo_data != NULL) {

        DBGPRINT(DBG_LOAD,
                 (KERN_DEBUG "%s: remove(fdo_data=%p)\n", __FUNCTION__,
                  fdo_data));

        DBGPRINT(DBG_LOAD,
                 (KERN_DEBUG
                  "%s: remove_card_callback(%p) (surprise_removed:%d)\n",
                  __FUNCTION__, sdio_remove_card_callback,
                  fdo_data->common.SurpriseRemoved));

                                                                                                                                                                                                /*--- mmoser 3/31/2006 ---*/
        if (fdo_data->common.SurpriseRemoved &&
            sdio_remove_card_callback != NULL) {
            sdio_remove_card_callback();
        }

        if (fdo_data->common.thread_id) {
            DBGPRINT(DBG_LOAD,
                     (KERN_DEBUG "remove() kill SDIOThread ... \n"));

            kill_proc(fdo_data->common.thread_id, SIGTERM, 1);

            // mmoser 2005-11-21
            fdo_data->common.stop_thread = SK_TRUE;

            wait_for_completion(&on_exit);
            DBGPRINT(DBG_LOAD,
                     (KERN_DEBUG "remove() SDIOThread killed ... \n"));
        }

                                                                                                                                                                                                /*--- mmoser 4/4/2006 ---*/
        SDHost_DisableInterrupt(&fdo_data->common,
                                STDHOST_NORMAL_IRQ_CARD_ALL_ENA);

        free_irq(fdo_data->common.Interrupt, &fdo_data->common);
        for (j = 0; j < fdo_data->common.IOBaseIx; j++)
            iounmap(fdo_data->common.IOBase[j]);
                                                                                                                                                                        /*--- mmoser 10/11/2006 ---*/
        pci_release_regions(fdo_data->common.dev);

    }
}

BOOLEAN
SDIOBus_CardRemoved(PFDO_DEVICE_DATA FdoData)
{

    ENTER();

    FdoData->CardIn = SK_FALSE;

    // mmoser 2005-11-29
    atomic_inc(&FdoData->common.card_remove_event);

    LEAVE();
    return SK_TRUE;
}

BOOLEAN
SDIOBus_CardPluggedIn(PFDO_DEVICE_DATA FdoData)
{
    SK_U32 ulVal;

    ENTER();
#ifdef USE_DEBOUNCE_CARD_IN

                                                                                                                                                                        /*--- mmoser 12/21/2006 ---*/
    mdelay(MAX_DEBOUNCE_TIME);

    // Read present state register to check whether a SD card is present
    MEM_READ_ULONG(&FdoData->common, SK_SLOT_0, STDHOST_PRESENT_STATE,
                   &ulVal);
    if ((ulVal & STDHOST_STATE_CARD_INSERTED) == 0) {
        // There is probably no card in the socket
        DBGPRINT(DBG_W528D,
                 (KERN_DEBUG
                  "SDHost_isCardPresent(): Card is NOT present. (%8.08X)\n",
                  ulVal));
        return SK_FALSE;
    }
#endif // USE_DEBOUNCE_CARD_IN

    SDHost_Init(&FdoData->common);

    if (!SDHost_InitializeCard(&FdoData->common)) {
        SDHost_EnableInterrupt((PCOMMON_DEVICE_DATA) & FdoData->common,
                               STDHOST_NORMAL_IRQ_CARD_ALL_ENA);
        return SK_FALSE;
    }
    FdoData->CardIn = SK_TRUE;

    DBGPRINT(DBG_LOAD,
             (KERN_DEBUG "Device Name : %s\n", FdoData->common.DeviceName));

    // mmoser 2005-11-29
    atomic_inc(&FdoData->common.card_add_event);
    LEAVE();
    return SK_TRUE;
}

static int
SDIOThread(void *data)
{
    PFDO_DEVICE_DATA fdo;
    SK_U16 usVal;
    SK_U32 card_in = 0;
    SK_U32 card_out = 0;
    PCOMMON_DEVICE_DATA pDev;

    daemonize("SDIOThread");
    allow_signal(SIGTERM);

    DBGPRINT(DBG_LOAD, (KERN_DEBUG "SDIOThread() started ... \n"));

    if (data == NULL) {
        DBGPRINT(DBG_LOAD, (KERN_DEBUG "SDIOThread() data == NULL !\n"));
        complete_and_exit(&on_exit, 0);
        return 1;
    }
    fdo = (PFDO_DEVICE_DATA) data;

    pDev = (PCOMMON_DEVICE_DATA) & fdo->common;
    LED_OFF(4);

    while (fdo != NULL && !fdo->common.stop_thread) {
        LED_ON(4);

// mmoser 2005-11-21
        // Read the Host Controller Version to check if SD Host adapter is available
        MEM_READ_USHORT(&fdo->common, SK_SLOT_0,
                        STDHOST_HOST_CONTROLLER_VERSION, &usVal);
        if (usVal == 0xFFFF) {
            // There is probably no SD Host Adapter available
            DBGPRINT(DBG_ERROR,
                     (KERN_DEBUG
                      "SDHost_isCardPresent(): SD Host Adapter is NOT present. (%4.04X)\n",
                      usVal));
            fdo->common.SurpriseRemoved = SK_TRUE;
            DBGPRINT(DBG_LOAD, (KERN_DEBUG "SDIOThread terminated ... \n"));
            complete_and_exit(&on_exit, 0);
            return -ENODEV;
        }
        // mmoser 2005-11-29
        card_in = atomic_read(&fdo->common.card_add_event);
        card_out = atomic_read(&fdo->common.card_remove_event);

        // mmoser 2005-11-29
        if (sdio_add_card_callback == NULL ||
            sdio_remove_card_callback == NULL ||
            (card_in == 0 && card_out == 0)) {
            LED_OFF(4);

                                                                                                                                                                        /*--- mmoser 6/21/2006 ---*/
            set_current_state(TASK_INTERRUPTIBLE);
            schedule_timeout(10);
            continue;
        }
        DBGPRINT(DBG_LOAD,
                 (KERN_DEBUG "%s: card_in=%d card_out=%d\n", __FUNCTION__,
                  card_in, card_out));

        // mmoser 2005-11-29
        if (card_in > 0) {
            DBGPRINT(DBG_LOAD,
                     (KERN_DEBUG "%s: CARD_ADD_EVENT received.\n",
                      __FUNCTION__));
            if (sdio_add_card_callback != NULL) {
                DBGPRINT(DBG_LOAD | DBG_ERROR,
                         (KERN_DEBUG
                          "%s: add_card_callback(0x%8.08X) SurpriseRemoved=0\n",
                          __FUNCTION__, (ULONG) fdo));
//                              pPriv = add_card_callback((PVOID)fdo) ;

                                                                                                                                                                        /*--- mmoser 4/3/2006 ---*/
                fdo->common.SurpriseRemoved = 0;
                                                                                                                                                                                                        /*--- mmoser 3/31/2006 ---*/
                sdio_add_card_callback(&pDev->dev->dev);
                // mmoser 2006-01-24
                // mmoser 2005-11-29
                atomic_dec(&fdo->common.card_add_event);

            } else {
                DBGPRINT(DBG_LOAD,
                         (KERN_DEBUG
                          "%s: add_card_callback(0x%8.08X) NOT INSTALLED !!!!!!!!!!!!!!\n",
                          __FUNCTION__, (ULONG) fdo));
            }
        }
        // mmoser 2005-11-29
        else if (card_out > 0) {
            DBGPRINT(DBG_LOAD,
                     (KERN_DEBUG "%s: CARD_REMOVE_EVENT received.\n",
                      __FUNCTION__));
            if (sdio_remove_card_callback != NULL) {
                DBGPRINT(DBG_LOAD | DBG_ERROR,
                         (KERN_DEBUG "%s: remove_card_callback(0x%8.08X)\n",
                          __FUNCTION__, (ULONG) fdo));

                                                                                                                                                                                                        /*--- mmoser 3/31/2006 ---*/
                sdio_remove_card_callback();
//                              remove_card_callback((PVOID)fdo);

                // mmoser 2006-01-24
                // mmoser 2005-11-29

                                                                                                                                                                                                        /*--- mmoser 3/31/2006 ---*/
                atomic_set(&fdo->common.card_remove_event, 0);
            } else {
                DBGPRINT(DBG_LOAD,
                         (KERN_DEBUG
                          "%s: remove_card_callback(0x%8.08X) NOT INSTALLED !!!!!!!!!!!!!!\n",
                          __FUNCTION__, (ULONG) fdo));
            }
        } else {
            DBGPRINT(DBG_LOAD, (KERN_DEBUG "SDIOThread terminated ... \n"));
            complete_and_exit(&on_exit, 0);
            return 1;
        }

    }

    DBGPRINT(DBG_LOAD, (KERN_DEBUG "SDIOThread terminated ... \n"));

    fdo->common.thread_id = 0;
    complete_and_exit(&on_exit, 0);
    return 0;
}

VOID
SDIOBus_InterfaceReference(IN PVOID Context)
{
    atomic_inc(&((PFDO_DEVICE_DATA) Context)->SDIOInterfaceRefCount);
}

VOID
SDIOBus_InterfaceDereference(IN PVOID Context)
{
    atomic_dec(&((PFDO_DEVICE_DATA) Context)->SDIOInterfaceRefCount);
}

void
mrvlsdio_register_callbacks(sdio_add_card_notifier add,
                            sdio_remove_card_notifier _remove)
{
    ENTER();

    DBGPRINT(DBG_LOAD,
             (KERN_DEBUG "%s: add=%p remove=%p\n", __FUNCTION__, add,
              _remove));

    sdio_add_card_callback = add;
    sdio_remove_card_callback = _remove;

    LEAVE();
}

int
mrvlsdio_get_interface(PGUID guid, ULONG size, ULONG version, PVOID interface)
{
    PMRVL_SDIO_INTERFACE_STANDARD SDIOInterfaceStandard;

    ENTER();

    if (guid == NULL || size == 0 || version == 0 || interface == NULL) {
        DBGPRINT(DBG_ERROR,
                 (KERN_DEBUG
                  "mrvlsdio_get_interface() called with illegal parameters!!!!\n"));
        return -1;
    }
    DBGPRINT(DBG_LOAD, (KERN_DEBUG "mrvlsdio_get_interface() called\n"));
    SHOW_GUID(*guid);
    if (COMPARE_GUID(*guid, GUID_MRVL_SDIOBUS_INTERFACE_2_STANDARD)) {
        DBGPRINT(DBG_LOAD,
                 (KERN_DEBUG "mrvlsdio_get_interface() interface is OK!\n"));

        if (size < sizeof(MRVL_SDIO_INTERFACE_STANDARD) && version != 1) {
            return -1;
        }

        if (fdo_data == NULL) {
            DBGPRINT(DBG_ERROR,
                     (KERN_DEBUG
                      "mrvlsdio_get_interface() fdo_data is NULL !!!!\n"));
            return -1;
        }

        SDIOInterfaceStandard = (PMRVL_SDIO_INTERFACE_STANDARD) interface;

        SDIOInterfaceStandard->Context = fdo_data;
        //
        // Fill in the exported functions
        //
        SDIOInterfaceStandard->InterfaceReference =
            (PINTERFACE_REFERENCE) SDIOBus_InterfaceReference;
        SDIOInterfaceStandard->InterfaceDereference =
            (PINTERFACE_DEREFERENCE) SDIOBus_InterfaceDereference;
        SDIOInterfaceStandard->SDIO_Cmd52_Read =
            (PSDIO_CMD52_READ) SDHost_CMD52_Read;
        SDIOInterfaceStandard->SDIO_Cmd52_Write =
            (PSDIO_CMD52_WRITE) SDHost_CMD52_Write;
        SDIOInterfaceStandard->SDIO_Cmd53_Read =
            (PSDIO_CMD53_READ) SDHost_CMD53_Read;
        SDIOInterfaceStandard->SDIO_Cmd53_Write =
            (PSDIO_CMD53_WRITE) SDHost_CMD53_Write;
        SDIOInterfaceStandard->SDIO_Cmd53_Read_Ex =
            (PSDIO_CMD53_READ_EX) SDHost_CMD53_ReadEx;
        SDIOInterfaceStandard->SDIO_Cmd53_Write_Ex =
            (PSDIO_CMD53_WRITE_EX) SDHost_CMD53_WriteEx;
        SDIOInterfaceStandard->SDIO_RegisterIsrHandler =
            (PSDIO_REGISTER_ISR_HANDLER) SDIO_RegisterIsrHandler;
        SDIOInterfaceStandard->SDIO_EnableInterrupt =
            (PSDIO_ENABLE_IRQ) SDHost_if_EnableInterrupt;
        SDIOInterfaceStandard->SDIO_DisableInterrupt =
            (PSDIO_DISABLE_IRQ) SDHost_if_DisableInterrupt;
        SDIOInterfaceStandard->SDIO_isCardPresent =
            (PSDIO_IS_CARD_PRESENT) SDHost_isCardPresent;
        SDIOInterfaceStandard->SDIO_lockInterface =
            (PSDIO_SDIO_LOCKINTERFACE) SDHost_lockInterface;
        SDIOInterfaceStandard->SDIO_unlockInterface =
            (PSDIO_SDIO_UNLOCKINTERFACE) SDHost_unlockInterface;
        SDIOInterfaceStandard->SDIO_SetLed = (PSDIO_SET_LED) SDHost_SetLed;
        SDIOInterfaceStandard->SDIO_PsState = (PSDIO_PS_STATE) SDHost_PsState;  //  pweber 17.08.2005

                                                                                                                                                                                                                                                                                                                                                                /*--- mmoser 12/18/2006 ---*/
        SDIOInterfaceStandard->SDIO_SetClock =
            (PSDIO_SET_CLOCK) SDHost_SetClock;
        SDIOInterfaceStandard->SDIO_SetBusWidth =
            (PSDIO_SET_BUSWIDTH) SDHost_SetBusWidth;
                                                                                                                                                                                                                                                                                                                                                                /*--- mmoser 1/2/2007 ---*/
        SDIOInterfaceStandard->SDIO_SetClockSpeed =
            (PSDIO_SET_CLOCK_SPEED) SDHost_SetClockSpeed;

                                                                                                                                                                                                                                                                                                                                                                /*--- mmoser 2/20/2007 ---*/
        SDIOInterfaceStandard->SDIO_SetGPO = (PSDIO_SET_GPO) SDHost_SetGPO;

        //
        // Must take a reference before returning
        //
        SDIOBus_InterfaceReference(fdo_data);

    }
    LEAVE();
    return 0;
}

void
SDIO_RegisterIsrHandler(PCOMMON_DEVICE_DATA pDev,
                        PVOID isrContext, PSDIO_ISR_HANDLER pIsrHandler)
{
    ENTER();

    pDev->isrContext = isrContext;
    pDev->pIsr = pIsrHandler;

    DBGPRINT(DBG_LOAD,
             (KERN_DEBUG
              "SDIO_RegisterIsrHandler() fdo=0x%8.08X ctx=0x%p handler=0x%8.08X\n",
              (ULONG) pDev, isrContext, (ULONG) pIsrHandler));
    LEAVE();
}

// for filling the "/proc/mrvsdio" entry
int
mrvlsdio_read_procmem(char *buf,
                      char **start,
                      off_t offset, int count, int *eof, void *data)
{
    int len = 0;

    if (fdo_data != NULL) {
        // after probe-event, we have "good" data in /proc/mrvsdio
        len =
            sprintf(buf,
                    "mrvlsdio is running with:\nbus_type     = 0x%x\nbus_with     = 0x%x\nclock_speed  = 0x%x\ndebug_flags  = 0x%x\nblock_size   = 0x%x\nsdio_voltage = 0x%x\n",
                    fdo_data->common.bus_type, fdo_data->common.bus_width,
                    fdo_data->common.ClockSpeed, fdo_data->common.debug_flags,
                    fdo_data->common.blockSize,
                    fdo_data->common.sdio_voltage);
    } else {
        // not probed yet
        len = sprintf(buf, "mrvlsdio is up and waiting for card\n");
    }

    *eof = 1;
    return (len);
}

static int
mrvlsdio_init_module(void)
{
    int ret;

    // generate /proc/ entry
    create_proc_read_entry("mrvlsdio",  // entry in "/proc/mrvlsdio"
                           0,   // file attributes
                           NULL,        // parent dir
                           mrvlsdio_read_procmem,       // name of function
                           NULL);       // client data

    DBGPRINT(DBG_ALL, ("*** mrvlsdio ***\n"));
    DBGPRINT(DBG_ALL, ("*** Built on %s %s ***\n", __DATE__, __TIME__));

                                                                                                                                                                        /*--- mmoser 4/4/2006 ---*/
    fdo_data =
        (PFDO_DEVICE_DATA) kmalloc(sizeof(FDO_DEVICE_DATA), GFP_KERNEL);

    if (fdo_data == NULL) {
        return -ENOMEM;
    }
    memset(fdo_data, 0, sizeof(FDO_DEVICE_DATA));

    DBGPRINT(DBG_LOAD, (KERN_DEBUG "Module mrvlsdio init: INIT_WORK\n"));

    INIT_WORK(&sdiobus_wq, (void (*)(void *)) SDIOBus_Dpc, NULL);

    ret = register_chrdev(sdio_major, "mrvlsdio", &mrvlsdio_fops);
    if (ret < 0) {
        DBGPRINT(DBG_LOAD,
                 (KERN_WARNING "mrvlsdio: can't get major %d  ret=%d\n",
                  sdio_major, ret));
    } else {
        DBGPRINT(DBG_LOAD, (KERN_DEBUG "mrvlsdio: major %d\n", ret));
    }
    if (sdio_major == 0)
        sdio_major = ret;       /* dynamic */

        /** from 2.4 kernel source:
	 * pci_register_driver - register a new pci driver
	 *
	 * Adds the driver structure to the list of registered drivers
	 * Returns the number of pci devices which were claimed by the driver
	 * during registration.  The driver remains registered even if the
	 * return value is zero.
	 */

    ret = pci_register_driver(&mrvlsdio);
    DBGPRINT(DBG_LOAD,
             (KERN_DEBUG "%s: pci_register_driver with %d\n", __FILE__, ret));

//      return (ret);
    return 0;
}

static void
mrvlsdio_exit_module(void)
{
    // remove proc entry
    remove_proc_entry("mrvlsdio", NULL);
    DBGPRINT(DBG_LOAD, (KERN_DEBUG "Module mrvlsdio exit\n"));

    unregister_chrdev(sdio_major, "mrvlsdio");
    pci_unregister_driver(&mrvlsdio);

                                                                                                                                                                                /*--- mmoser 4/4/2006 ---*/
    kfree(fdo_data);

    fdo_data = NULL;
}

module_init(mrvlsdio_init_module);
module_exit(mrvlsdio_exit_module);

EXPORT_SYMBOL(mrvlsdio_get_interface);
EXPORT_SYMBOL(mrvlsdio_register_callbacks);

#endif
